/* this file is a part of amp software

   buffer.c: written by Andrew Richards  <A.Richards@phys.canterbury.ac.nz>
             (except printout())

   Last modified by:
   Karl Anders Oygard added flushing of audio buffers, 13 May 1997
   Isaac Richards, emp3 support, 6 November 1997   
   Isaac Richards, threaded buffer, using pthreads, 19 December 1997
*/
#include "audio.h"


#include <sys/types.h>
#include <signal.h>
#ifdef HAVE_MLOCK
#ifdef OS_Linux
#include <sys/mman.h>
#endif
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xthreads.h>
#include "../analyze.h"

struct audioType {
        int frequency;
        int stereo;
        int volume;
};

int notadjusted;
pthread_mutex_t  buffer_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_t buffer_thread;

struct ringBuffer audioBuffer;

int buffer_running;
int doFlush;
int inputFinished;
int AUDIO_BUFFER_SIZE;

void audioThread(struct audioType *type);

/* little endian systems do not require any byte swapping whatsoever. 
 * big endian systems require byte swapping when writing .wav files, 
 * but not when playing directly
 */

int
audioBufferWrite(char *buf,int bytes)
{
   if (notadjusted && (AUDIO_BUFFER_SIZE % bytes) != 0) {
	AUDIO_BUFFER_SIZE = (AUDIO_BUFFER_SIZE / bytes) * bytes;
	notadjusted = FALSE;
   } 
   while ((!audioBuffer.bufferPtr) || (bufferFree(&audioBuffer) < bytes))
	usleep(500);
   if (!audioBuffer.bufferPtr) {
      fprintf(stderr,"nuthin to write to\n");
      return 0;
   }
   pthread_mutex_lock(&buffer_mutex);
   memcpy((audioBuffer.bufferPtr + audioBuffer.inPos), buf, bytes);
   audioBuffer.inPos = (audioBuffer.inPos + bytes) % AUDIO_BUFFER_SIZE;
   pthread_mutex_unlock(&buffer_mutex);
   return bytes;
}

static inline int writeAudio(char *buffer, int size)
{
   if (use_wave) {
#ifndef NO_BYTE_SWAPPING
      short *ptr = (short*) buffer;
      int i = size / sizeof(short);
      for (;i>=0;--i)
         ptr[i] = ptr[i] << 8 | ptr[i] >> 8;
#endif
      return fwrite(buffer, sizeof(char), size, out_file);
   }
   else 
      return audioWrite(buffer,size);
}

int audio_play(char *buffer, int size)
{
   if (AUDIO_BUFFER_SIZE==0)
      return writeAudio(buffer, size);
   else 
      return audioBufferWrite(buffer, size);
}

void 
initBuffer(void)
{
	pthread_mutex_lock(&buffer_mutex);
	audioBuffer.bufferPtr=malloc(AUDIO_BUFFER_SIZE+20);
	if (audioBuffer.bufferPtr==NULL) {
		fprintf(stderr,"Unable to allocate write buffer\n");
		exit(1);
	}
	memset(audioBuffer.bufferPtr,0,AUDIO_BUFFER_SIZE);
#ifdef HAVE_MLOCK
	mlock(audioBuffer.bufferPtr,AUDIO_BUFFER_SIZE);
#endif
	audioBuffer.inPos = 0;
	audioBuffer.outPos = 0;
	pthread_mutex_unlock(&buffer_mutex);
}

void
freeBuffer(void)
{
	pthread_mutex_lock(&buffer_mutex);
#ifdef HAVE_MLOCK
        munlock(audioBuffer.bufferPtr,AUDIO_BUFFER_SIZE);
#endif
	free(audioBuffer.bufferPtr);
	audioBuffer.bufferPtr = NULL;
	pthread_mutex_unlock(&buffer_mutex);
}

void
audioBufferFlush()
{
	if (AUDIO_BUFFER_SIZE!=0)
	        doFlush = TRUE;
	else
		audioFlush();
}
	
/* The decoded data are stored in a the ring buffer until they can be sent */
/* to the audio device. Variables are named in relation to the buffer */
/* ie writes are writes from the codec to the buffer and reads are reads */
/* from the buffer to the audio device */

int
audioBufferOpen(int frequency, int stereo, int volume)
{
        struct audioType *type = malloc(sizeof(struct audioType));

	inputFinished = FALSE;
	notadjusted = TRUE;
	audioBuffer.bufferPtr = NULL;
	
	type->frequency = frequency;
	type->stereo = stereo;
	type->volume = volume;
        pthread_create(&buffer_thread, NULL, (void *) audioThread, (void *)type);
        sleep(1);	
	return 1;
}
	
void audioThread(struct audioType *type)
{
        int outFd, cnt /* , cntr = 0 */;
	int percentFull, bytesToEnd;
	fd_set outFdSet;
	fd_set *outFdPtr;
	struct timeval timeout;

	if (audioOpen(type->frequency,type->stereo,type->volume)) {
	   fprintf(stderr,"Urghggghghghg... the buffer dies a horrible death...\n");
	   exit(0);
	}

	outFd=getAudioFd();

        initBuffer();

	free(type);

	buffer_running = TRUE;
	while(1) {
		usleep(10);
		FD_ZERO(&outFdSet);
		FD_SET(outFd, &outFdSet);
/*			 fprintf(stderr,"BS=%d inPos=%d outPos=%d bufferSize=%d bufferFree=%d\n",
					 AUDIO_BUFFER_SIZE,audioBuffer.inPos,audioBuffer.outPos,
					 bufferSize(&audioBuffer),bufferFree(&audioBuffer));
*/
		pthread_mutex_lock(&buffer_mutex);
		if (bufferSize(&audioBuffer) < AUSIZ) {	/* is the buffer too empty */
			outFdPtr=NULL;		/* yes, don't try to write */
			if (inputFinished) {	/* no more input, buffer exhausted -> exit */
				pthread_mutex_unlock(&buffer_mutex);
				break;
			}
		} else
			outFdPtr=&outFdSet;	
		pthread_mutex_unlock(&buffer_mutex);			      			
		/* check if we can write */
		if (select(outFd+1,NULL,outFdPtr,NULL,&timeout) > -1) { 
			if (outFdPtr && FD_ISSET(outFd,outFdPtr)) {	 						/* need to write */
				pthread_mutex_lock(&buffer_mutex);
				bytesToEnd = AUDIO_BUFFER_SIZE - audioBuffer.outPos;

				percentFull=100*bufferSize(&audioBuffer)/AUDIO_BUFFER_SIZE;
/*				if ((cntr++ % (16/(AUSIZ/4096)))==0) fprintf(stderr,"\rBuffer (%2d%%) %6d",percentFull,bufferSize(&audioBuffer)); 
*/				if (AUSIZ>bytesToEnd) {
					cnt = writeAudio(audioBuffer.bufferPtr + audioBuffer.outPos, bytesToEnd);
					cnt += writeAudio(audioBuffer.bufferPtr, AUSIZ - bytesToEnd);
					audioBuffer.outPos = AUSIZ - bytesToEnd;
				} else {
					cnt = writeAudio(audioBuffer.bufferPtr + audioBuffer.outPos, AUSIZ);
					audioBuffer.outPos += AUSIZ;
				}
				pthread_mutex_unlock(&buffer_mutex);
/*				fprintf(stderr,"buffer: wrote %d bytes (to audio)\n",cnt);
	*/		}

		} else {
		        perror("select");
			exit(-1); 
		}  
		if (doFlush) {
		        pthread_mutex_lock(&buffer_mutex);
		        audioBuffer.inPos = audioBuffer.outPos = 0;
		        memset(audioBuffer.bufferPtr,0, AUDIO_BUFFER_SIZE);
			audioFlush();
			doFlush = FALSE;
			pthread_mutex_unlock(&buffer_mutex);
		}
	}
        freeBuffer();
	audioClose();
	buffer_running = FALSE;
}

void
audioBufferClose()
{
   /* Close the buffer pipe thread */
   if (!end_of_song) {
      if (AUDIO_BUFFER_SIZE!=0) 
         doFlush = TRUE; 
       else 
         audioFlush();
   }
}
